@@ -31,33 +31,6 @@ cfile_readbuf(struct cfile *f, unsigned char *buf, int len)
       f->bufN = 0;
       return 0;
     }
-
-  if (f->nunread)
-    {
-      int ur = len >= f->nunread ? f->nunread : len;
-      memcpy(buf, f->unreadbuf, ur);
-      if (f->nunread > ur)
-	memmove(f->unreadbuf, f->unreadbuf + ur, f->nunread - ur);
-      f->nunread -= ur;
-      if (f->nunread == 0)
-	{
-	  if (f->unreadbuf != f->buf)
-	    free(f->unreadbuf);
-	  f->unreadbuf = 0;
-	}
-      if (f->len != CFILE_LEN_UNLIMITED)
-	f->len -= ur;
-      if (len > ur)
-	{
-	  int r = cfile_readbuf(f, buf + ur, len - ur);
-	  if (r == -1)
-	    return -1;
-	  ur += r;
-	}
-      f->bufN = ur;
-      return ur;
-    }
-
   switch (f->fd)
     {
     case CFILE_IO_FILE:
@@ -73,6 +46,16 @@ cfile_readbuf(struct cfile *f, unsigned char *buf, int len)
     case CFILE_IO_CFILE:
       len = ((struct cfile *)f->fp)->read((struct cfile *)f->fp, buf, len);
       break;
+    case CFILE_IO_PUSHBACK:
+      len = ((struct cfile *)f->fp)->read((struct cfile *)f->fp, buf, len);
+      if (((struct cfile *)f->fp)->nunread == 0)
+	{
+	  struct cfile *cf = (struct cfile *)f->fp;
+	  f->fp = cf->fp;
+	  f->fd = cf->fd;
+	  cf->close(cf);
+	}
+      break;
     case CFILE_IO_ALLOC:
       return -1;
     case CFILE_IO_BUFFER:
@@ -124,9 +107,13 @@ cfile_writebuf(struct cfile *f, unsigned char *buf, int len)
       break;
     case CFILE_IO_ALLOC:
       bp = (unsigned char **)f->fp;
+      if (f->bytes + len < f->bytes)
+	return -1;
       if (!f->bytes || (((f->bytes + len - 1) ^ (f->bytes - 1)) & ~0x1fff) != 0)
 	{
 	  int ns = (len + f->bytes + 0x1fff) & ~0x1fff;
+	  if (ns < f->bytes + len)
+	    return -1;
 	  if (!f->bytes)
 	    nb = malloc(ns);
 	  else
@@ -138,13 +125,12 @@ cfile_writebuf(struct cfile *f, unsigned char *buf, int len)
       memcpy(*bp + f->bytes, buf, len);
       break;
     case CFILE_IO_NULL:
-      return -1;
       break;
     default:
       len = write(f->fd, buf, len);
-      if (len == -1)
-        return -1;
     }
+  if (len == -1)
+    return -1;
   if (f->len != CFILE_LEN_UNLIMITED)
     f->len -= len;
   if (len && f->ctxup)
@@ -153,6 +139,56 @@ cfile_writebuf(struct cfile *f, unsigned char *buf, int len)
   return len;
 }
 
+static void
+cwclose_fixupalloc(struct cfile *f)
+{
+  unsigned char *n, **bp = (unsigned char **)f->fp;
+  n = *bp;
+  if (!n)
+    return;
+  n = realloc(n, f->bytes);
+  if (n)
+    *bp = n;
+}
+
+
+/*****************************************************************
+ *  unread stuff
+ */
+
+static int
+crread_ur(struct cfile *f, void *buf, int len)
+{
+  int l2;
+  l2 = len > f->nunread ? f->nunread : len;
+  if (l2)
+    {
+      memcpy(buf, f->unreadbuf, l2);
+      buf += l2;
+      len -= l2;
+      f->nunread -= l2;
+      if (f->ctxup)
+	f->ctxup(f->ctx, f->unreadbuf, l2);
+      f->bytes += l2;
+      if (f->nunread)
+	memmove(f->unreadbuf, f->unreadbuf + l2, f->nunread);
+      if (!f->nunread && f->unreadbuf != f->buf)
+	{
+	  free(f->unreadbuf);
+	  f->unreadbuf = 0;
+	}
+    }
+  if (!f->nunread)
+    {
+      f->read = f->oldread;
+      f->oldread = 0;
+    }
+  if (!len)
+    return l2;
+  len = f->read(f, buf, len);
+  return len == -1 ? -1 : l2 + len;
+}
+
 static int
 cfile_unreadbuf(struct cfile *f, void *buf, int len, int usebuf)
 {
@@ -179,26 +215,18 @@ cfile_unreadbuf(struct cfile *f, void *buf, int len, int usebuf)
 	return -1;
     }
   if (f->nunread)
-    memmove(newbuf, newbuf + len, f->nunread);
+    memmove(newbuf + len, newbuf, f->nunread);
   memcpy(newbuf, buf, len);
   f->unreadbuf = newbuf;
   f->nunread += len;
-  if (f->len != CFILE_LEN_UNLIMITED)
-    f->len += len;
+  if (f->read != crread_ur)
+    {
+      f->oldread = f->read;
+      f->read = crread_ur;
+    }
   return 0;
 }
 
-static void
-cwclose_fixupalloc(struct cfile *f)
-{
-  unsigned char *n, **bp = (unsigned char **)f->fp;
-  n = *bp;
-  if (!n)
-    return;
-  n = realloc(n, f->bytes);
-  if (n)
-    *bp = n;
-}
 
 /*****************************************************************
  *  bzip2 io
@@ -420,6 +448,11 @@ crclose_gz(struct cfile *f)
       if (cf->unread(cf, f->strm.gz.next_in, f->strm.gz.avail_in) != -1)
         f->strm.gz.avail_in = 0;
     }
+  if (f->fd == CFILE_IO_PUSHBACK)
+    {
+      struct cfile *cf = (struct cfile *)f->fp;
+      cf->close(cf);
+    }
   r = (f->len != CFILE_LEN_UNLIMITED ? f->len : 0) + f->strm.gz.avail_in;
   if (f->unreadbuf != f->buf)
     free(f->unreadbuf);
@@ -632,10 +665,13 @@ crclose_un(struct cfile *f)
 static struct cfile *
 cropen_un(struct cfile *f)
 {
-  if (f->bufN != -1)
+  if (f->bufN != -1 && f->bufN != 0)
     {
+      /* CFILE_COMP_XX read some bytes, set up unread */
       f->unreadbuf = f->buf;
       f->nunread = f->bufN;
+      f->oldread = f->read;
+      f->read = crread_ur;
     }
   return f;
 }
@@ -798,10 +834,34 @@ cfile_detect_rsync(struct cfile *f)
     f->comp = comp;
   if (len)
     {
+      struct cfile *cf;
+      if (f->fd == CFILE_IO_CFILE || f->fd == CFILE_IO_PUSHBACK)
+	{
+	  cf = (struct cfile *)f->fp;
+	  if (!cf->unread(cf, b, len))
+	    {
+	      free(b);
+	      return -1;
+	    }
+	  free(b);
+	}
+      else
+	{
+	  cf = cfile_open(CFILE_OPEN_RD, f->fd, f->fp, CFILE_COMP_UN, CFILE_LEN_UNLIMITED, 0, 0);
+	  if (!cf)
+	    {
+	      free(b);
+	      return -1;
+	    }
+	  f->fp = cf;
+	  f->fd = CFILE_IO_PUSHBACK;
+	  cf->unreadbuf = b;
+	  cf->nunread = len;
+	  cf->oldread = cf->read;
+	  cf->read = crread_ur;
+	}
       if (f->len != CFILE_LEN_UNLIMITED)
 	f->len += len;
-      f->nunread = len;
-      f->unreadbuf = b;
     }
   else
     free(b);
@@ -813,14 +873,21 @@ cfile_detect_rsync(struct cfile *f)
  */
 
 struct cfile *
-cfile_open(int mode, int fd, void *fp, int comp, unsigned int len, void (*ctxup)(void *, unsigned char *, unsigned int), void *ctx)
+cfile_open(int mode, int fd, void *fp, int comp, size_t len, void (*ctxup)(void *, unsigned char *, unsigned int), void *ctx)
 {
   struct cfile *f;
   if (comp == CFILE_COMP_XX && mode == CFILE_OPEN_WR)
     return 0;
   if (mode != CFILE_OPEN_RD && mode != CFILE_OPEN_WR)
     return 0;
-  f = malloc(sizeof(*f));
+  if (fd == CFILE_IO_REOPEN)
+    {
+      f = fp;
+      fd = f->fd;
+      fp = f->fp;
+    }
+  else
+    f = malloc(sizeof(*f));
   if (!f)
     return 0;
   f->fd = fd;
@@ -832,6 +899,7 @@ cfile_open(int mode, int fd, void *fp, int comp, unsigned int len, void (*ctxup)
   f->bufN = -1;
   f->nunread = 0;
   f->unreadbuf = 0;
+  f->oldread = 0;
   if (mode == CFILE_OPEN_WR && fd == CFILE_IO_ALLOC)
     {
       unsigned char **bp = (unsigned char **)f->fp;
@@ -840,7 +908,7 @@ cfile_open(int mode, int fd, void *fp, int comp, unsigned int len, void (*ctxup)
   if (comp == CFILE_COMP_XX)
     {
       comp = CFILE_COMP_UN;
-      if (len == -1 || len >= 2)
+      if (len == CFILE_LEN_UNLIMITED || len >= 2)
 	{
 	  int n = cfile_readbuf(f, f->buf, sizeof(f->buf));
 	  if (n == -1)
