@@ -1,15 +1,37 @@
 /*
   diff.c -- binary diff generator
 
-  Copyright 2003 Colin Percival
-  Copyright 2004 Michael Schroeder
+  Copyright 2004,2005 Michael Schroeder
 
-  rewritten from bsdiff.c, added library interface and hash method,
-  enhanced suffix method.
-
-  For the terms under which this work may be distributed, please see
-  the adjoining file "LICENSE.BSDPL".
+  rewritten from bsdiff.c,
+      http://www.freebsd.org/cgi/cvsweb.cgi/src/usr.bin/bsdiff
+  added library interface and hash method, enhanced suffix method.
 */
+/*-
+ * Copyright 2003-2005 Colin Percival
+ * All rights reserved
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted providing that the following conditions 
+ * are met:
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer.
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notice, this list of conditions and the following disclaimer in the
+ *    documentation and/or other materials provided with the distribution.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
+ * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
+ * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
+ * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
+ * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
+ * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
+ * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
+ * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
+ * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
+ * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+ * POSSIBILITY OF SUCH DAMAGE.
+ */
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -62,6 +84,8 @@ static int blockwrite(struct bzblock *bzf, void *buf, int len)
     {
       if (bzf->strm.avail_out < 4096)
 	{
+	  if (bzf->len + 8192 < bzf->len)
+	    return -1;
 	  if (bzf->data)
 	    bzf->data = realloc(bzf->data, bzf->len + 8192);
 	  else
@@ -80,7 +104,7 @@ static int blockwrite(struct bzblock *bzf, void *buf, int len)
     }
 }
 
-static int blockclose(struct bzblock *bzf, unsigned char **datap, int *lenp)
+static int blockclose(struct bzblock *bzf, unsigned char **datap, unsigned int *lenp)
 {
   int ret;
   bzf->strm.avail_in = 0;
@@ -89,6 +113,8 @@ static int blockclose(struct bzblock *bzf, unsigned char **datap, int *lenp)
     {
       if (bzf->strm.avail_out < 4096)
 	{
+	  if (bzf->len + 8192 < bzf->len)
+	    return -1;
 	  if (bzf->data)
 	    bzf->data = realloc(bzf->data, bzf->len + 8192);
 	  else
@@ -112,10 +138,10 @@ static int blockclose(struct bzblock *bzf, unsigned char **datap, int *lenp)
   return 0;
 }
 
-static inline unsigned int matchlen(unsigned char *old, int oldlen, unsigned char *new, int newlen)
+static inline bsuint matchlen(unsigned char *old, bsuint oldlen, unsigned char *new, bsuint newlen)
 {
   unsigned char *oldsave;
-  int max;
+  bsuint max;
   oldsave = old;
   max = oldlen > newlen ? newlen : oldlen;
   while (max-- > 0)
@@ -124,6 +150,7 @@ static inline unsigned int matchlen(unsigned char *old, int oldlen, unsigned cha
   return old - oldsave;
 }
 
+#ifndef BSDIFF_NO_HASH
 
 /******************************************************************/
 /*                                                                */
@@ -209,26 +236,34 @@ static unsigned int primes[] =
   65537, 98317, 147481, 221227, 331841, 497771, 746659, 1120001,
   1680013, 2520031, 3780053, 5670089, 8505137, 12757739, 19136609,
   28704913, 43057369, 64586087, 96879131, 145318741, 217978121,
-  326967209, 490450837, 735676303, 1103514463, 1655271719
+  326967209, 490450837, 735676303, 1103514463, 1655271719,
+  0xffffffff
 };
 
 struct hash_data {
-  unsigned int *hash;
-  int prime;
+  bsuint *hash;
+  unsigned int prime;
 };
 
-static void *hash_create(unsigned char *buf, int len)
+static void *hash_create(unsigned char *buf, bsuint len)
 {
   struct hash_data *hd;
-  unsigned int *hash;
+  bsuint *hash;
   unsigned char *bp = buf;
-  unsigned int off, s;
+  bsuint off;
+  unsigned int s;
   unsigned int prime;
+  unsigned int num;
 
   hd = malloc(sizeof(*hd));
   if (!hd)
     return 0;
-  int num = (len + HSIZE - 1) >> HSIZESHIFT;
+#ifdef BSDIFF_64BIT
+  /* this is a 16GB limit for HSIZESHIFT == 4 */
+  if (len >= (bsuint)(0xffffffff / 4) << HSIZESHIFT)
+    return 0;
+#endif
+  num = (len + HSIZE - 1) >> HSIZESHIFT;
   prime = num * 4;
   for (s = 0; s < sizeof(primes)/sizeof(*primes) - 1; s++)
     if (prime < primes[s])
@@ -265,16 +300,16 @@ static void hash_free(void *data)
   free(hd);
 }
 
-static unsigned int hash_findnext(void *data, unsigned char *old, unsigned int oldlen, unsigned char *new, unsigned int newlen, unsigned int lastoffset, unsigned int scan, unsigned int *posp, unsigned int *lenp)
+static bsuint hash_findnext(void *data, unsigned char *old, bsuint oldlen, unsigned char *new, bsuint newlen, bsuint lastoffset, bsuint scan, bsuint *posp, bsuint *lenp)
 {
   struct hash_data *hd = data;
-  unsigned int scanstart, oldscore, oldscorenum, oldscorestart;
+  bsuint scanstart, oldscore, oldscorenum, oldscorestart;
+  bsuint pos, len;
+  bsuint lscan, lpos, llen;
+  bsuint i, ss, scsc;
   unsigned int ssx;
-  unsigned int pos, len;
-  unsigned int lscan, lpos, llen;
-  unsigned int i, ss, scsc;
-  int prime;
-  unsigned int *hash;
+  unsigned int prime;
+  bsuint *hash;
 
   hash = hd->hash;
   prime = hd->prime;
@@ -308,8 +343,6 @@ scannext:
 	  ssx ^= oldc;
 #endif
 	  scan++;
-	  if (ssx != buzhash(new + scan))
-	    abort();
 	  continue;
 	}
       pos--;
@@ -322,18 +355,19 @@ scannext:
 	  if (memcmp(old + pos, new + scan, HSIZE))
 	    goto scannext;
 	}
-      len = matchlen(old + pos + HSIZE,oldlen - pos - HSIZE,new + scan + HSIZE,newlen - scan - HSIZE) + HSIZE;
+      len = matchlen(old + pos + HSIZE, oldlen - pos - HSIZE, new + scan + HSIZE, newlen - scan - HSIZE) + HSIZE;
       if (scan + HSIZE * 4 <= newlen)
 	{
-	  unsigned int ss2, len2, pos2;
-	  ss2 = buzhash(new + scan + HSIZE * 3) % prime;
-	  pos2 = hash[ss2];
+	  unsigned int ssx2;
+	  bsuint len2, pos2;
+	  ssx2 = buzhash(new + scan + HSIZE * 3) % prime;
+	  pos2 = hash[ssx2];
 	  if (pos2)
 	    {
 	      if (memcmp(new + scan + HSIZE *3, old + pos2 - 1, HSIZE))
 		{
-		  ss2 = (ss2 == prime) ? 0 : ss2 + 1;
-		  pos2 = hash[ss2];
+		  ssx2 = (ssx2 == prime) ? 0 : ssx2 + 1;
+		  pos2 = hash[ssx2];
 		}
 	    }
 	  if (pos2 > 1 + HSIZE*3)
@@ -341,7 +375,7 @@ scannext:
 	      pos2 = pos2 - 1 - HSIZE*3;
 	      if (pos2 != pos)
 		{
-		  len2 = matchlen(old + pos2,oldlen - pos2,new + scan,newlen - scan);
+		  len2 = matchlen(old + pos2, oldlen - pos2, new + scan, newlen - scan);
 		  if (len2 > len)
 		    {
 		      pos = pos2;
@@ -420,6 +454,9 @@ gotit:
   return scan;
 }
 
+#endif /* BSDIFF_NO_HASH */
+
+#ifndef BSDIFF_NO_SUF
 
 /******************************************************************/
 /*                                                                */
@@ -428,13 +465,13 @@ gotit:
 /******************************************************************/
 
 struct suf_data {
-  int *I;
-  int F[257];
+  bsint *I;
+  bsint F[257];
 };
 
-static void suf_split(int *I, int *V, int start, int len, int h)
+static void suf_split(bsint *I, bsint *V, bsint start, bsint len, bsint h)
 {
-  int i, j, k, x, tmp, jj, kk;
+  bsint i, j, k, x, tmp, jj, kk;
 
   if (len < 16)
     {
@@ -529,12 +566,12 @@ static void suf_split(int *I, int *V, int start, int len, int h)
     suf_split(I, V, kk, start + len - kk, h);
 }
 
-static int suf_bucketsort(int *V, int *I, int n, int s)
+static bsint suf_bucketsort(bsint *V, bsint *I, bsint n, bsint s)
 {
-  int c, d, g, i, j;
-  int *B; 
+  bsint c, d, g, i, j;
+  bsint *B; 
 
-  B = calloc(s, sizeof(int));
+  B = calloc(s, sizeof(bsint));
   if (!B)
     return 0;
   for (i = n - 1; i >= 0; i--)
@@ -559,22 +596,25 @@ static int suf_bucketsort(int *V, int *I, int n, int s)
   return 1;
 }
 
-static void *suf_create(unsigned char *buf, int len)
+static void *suf_create(unsigned char *buf, bsuint ulen)
 {
   struct suf_data *sd;
-  int *V, *I;
-  int i, h, l, oldv, s;
+  bsint *V, *I;
+  bsint i, h, l, oldv, s, len;
 
+  len = ulen;
+  if (len < 0)
+    return 0;
   sd = malloc(sizeof(*sd));
   if (!sd)
     return 0;
-  V = malloc(sizeof(int) * (len + 3));
+  V = malloc(sizeof(bsint) * (len + 3));
   if (!V)
     {
       free(sd);
       return 0;
     }
-  I = malloc(sizeof(int) * (len + 3));
+  I = malloc(sizeof(bsint) * (len + 3));
   if (!I)
     {
       free(V);
@@ -671,9 +711,9 @@ static void suf_free(void *data)
   free(sd);
 }
 
-static unsigned int suf_bsearch(int *I, unsigned char *old, unsigned int oldlen, unsigned char *new, unsigned int newlen, unsigned int st, unsigned int en, unsigned int *posp)
+static bsuint suf_bsearch(bsint *I, unsigned char *old, bsuint oldlen, unsigned char *new, bsuint newlen, bsuint st, bsuint en, bsuint *posp)
 {
-  unsigned int x, y;
+  bsuint x, y;
   if (st > en)
     return 0;
   while (en - st >= 2)
@@ -690,9 +730,9 @@ static unsigned int suf_bsearch(int *I, unsigned char *old, unsigned int oldlen,
   return x > y ? x : y;
 }
 
-static unsigned int suf_findnext(void *data, unsigned char *old, unsigned int oldlen, unsigned char *new, unsigned int newlen, unsigned int lastoffset, unsigned int scan, unsigned int *posp, unsigned int *lenp)
+static bsuint suf_findnext(void *data, unsigned char *old, bsuint oldlen, unsigned char *new, bsuint newlen, bsuint lastoffset, bsuint scan, bsuint *posp, bsuint *lenp)
 {
-  unsigned int scsc, oldscore, len;
+  bsuint scsc, oldscore, len;
 
   struct suf_data *sd = data;
   len = 0;
@@ -722,23 +762,28 @@ static unsigned int suf_findnext(void *data, unsigned char *old, unsigned int ol
   return scan;
 }
 
+#endif /* BSDIFF_NO_SUF */
 
 /******************************************************************/
 
 struct deltamode {
   int mode;
-  void *(*create)(unsigned char *buf, int len);
-  unsigned int (*findnext)(void *data, unsigned char *old, unsigned int oldlen, unsigned char *new, unsigned int newlen, unsigned int lastoffset, unsigned int scan, unsigned int *posp, unsigned int *lenp);
+  void *(*create)(unsigned char *buf, bsuint len);
+  bsuint (*findnext)(void *data, unsigned char *old, bsuint oldlen, unsigned char *new, bsuint newlen, bsuint lastoffset, bsuint scan, bsuint *posp, bsuint *lenp);
   void (*free)(void *data);
 };
 
 struct deltamode deltamodes[] =
 {
+#ifndef BSDIFF_NO_SUF
   {DELTAMODE_SUF, suf_create, suf_findnext, suf_free}, 
+#endif
+#ifndef BSDIFF_NO_HASH
   {DELTAMODE_HASH, hash_create, hash_findnext, hash_free},
+#endif
 };
 
-static void addoff(struct bzblock *bzi, int off)
+static void addoff(struct bzblock *bzi, bsint off)
 {
   int i, sign = 0;
   unsigned char b[8];
@@ -757,21 +802,26 @@ static void addoff(struct bzblock *bzi, int off)
   blockwrite(bzi, b, 8);
 }
 
-void mkdiff(int mode, unsigned char *old, unsigned int oldlen, unsigned char *new, int newlen, struct instr **instrp, int *instrlenp, unsigned char **instrblkp, int *instrblklenp, unsigned char **addblkp, int *addblklenp, unsigned char **extrablkp, int *extrablklenp)
+void mkdiff(int mode, unsigned char *old, bsuint oldlen, unsigned char *new, bsuint newlen, struct instr **instrp, int *instrlenp, unsigned char **instrblkp, unsigned int *instrblklenp, unsigned char **addblkp, unsigned int *addblklenp, unsigned char **extrablkp, unsigned int *extrablklenp)
 {
   struct instr *instr = 0;
   int instrlen = 0;
-  unsigned int scan, pos, len;
-  unsigned int lastscan, lastpos, lastoffset;
-  int s, Sf, lenf, Sb, lenb;
-  int overlap, Ss, lens;
-  struct bzblock *bza;
+  bsuint i, scan, pos, len;
+  bsuint lastscan, lastpos, lastoffset;
+  bsuint s, Sf, lenf, Sb, lenb;
+  bsuint overlap, Ss, lens;
+  struct bzblock *bza = 0;
   struct bzblock *bze = 0;
   struct bzblock *bzi = 0;
-  int i;
   void *data;
   struct deltamode *dm;
+  int noaddblk = 0;
 
+  if ((mode & DELTAMODE_NOADDBLK) != 0)
+    {
+      mode ^= DELTAMODE_NOADDBLK;
+      noaddblk = 1;
+    }
   dm = 0;
   for (i = 0; i < sizeof(deltamodes)/sizeof(*deltamodes); i++)
     {
@@ -784,8 +834,11 @@ void mkdiff(int mode, unsigned char *old, unsigned int oldlen, unsigned char *ne
       fprintf(stderr, "mkdiff: no mode installed\n");
       exit(1);
     }
-  *addblkp = 0;
-  *addblklenp = 0;
+  if (addblkp)
+    {
+      *addblkp = 0;
+      *addblklenp = 0;
+    }
   if (extrablkp)
     {
       *extrablkp = 0;
@@ -796,7 +849,7 @@ void mkdiff(int mode, unsigned char *old, unsigned int oldlen, unsigned char *ne
       *instrblkp = 0;
       *instrblklenp = 0;
     }
-  if ((bza = blockopen()) == 0)
+  if (!noaddblk && addblkp && (bza = blockopen()) == 0)
     {
       fprintf(stderr, "mkdiff: could not create compression stream\n");
       exit(1);
@@ -818,52 +871,67 @@ void mkdiff(int mode, unsigned char *old, unsigned int oldlen, unsigned char *ne
       exit(1);
     }
 
-  scan = 0; len = 0;
+  scan = 0; len = 0; lenf = 0;
   lastscan = 0; lastpos = 0; lastoffset = 0;
+
   while (scan < newlen)
     {
+      if (noaddblk)
+	{
+	  lastoffset = oldlen;
+	  lenf = len;
+	}
       scan += len;
       scan = dm->findnext(data, old, oldlen, new, newlen, lastoffset, scan, &pos, &len);
-      s = Sf = lenf = 0;
-      for(i = 0; lastscan + i < scan && lastpos + i < oldlen; )
+      if (!noaddblk)
 	{
-	  if (old[lastpos+i] == new[lastscan+i])
-	    s++;
-	  i++;
-	  if (s * 2 - i >= Sf * 2 - lenf)
+	  s = Sf = lenf = 0;
+	  for(i = 0; lastscan + i < scan && lastpos + i < oldlen; )
 	    {
-	      Sf = s;
-	      lenf = i;
+	      if (old[lastpos+i] == new[lastscan+i])
+		{
+		  s++;
+		  i++;
+		  if (s >= Sf + i - s)
+		    {
+		      Sf = 2 * s - i;
+		      lenf = i;
+		    }
+		}
+	      else
+	        i++;
 	    }
 	}
       lenb = 0;
-      if (scan < newlen)
+      if (!noaddblk && scan < newlen)
 	{
 	  s = Sb = 0;
 	  for(i = 1; scan >= lastscan + i && pos >= i; i++)
 	    {
 	      if (old[pos-i] == new[scan-i])
-	        s++;
-	      if(s * 2 - i >= Sb * 2 - lenb)
 		{
-		  Sb = s;
-		  lenb = i;
+		  s++;
+		  if(s >= Sb + i - s)
+		    {
+		      Sb = 2 * s - i;
+		      lenb = i;
+		    }
 		}
 	    }
 	}
       if (lastscan + lenf > scan - lenb)
 	{
 	  overlap = (lastscan + lenf) - (scan - lenb);
-	  s = Ss = lens = 0;
+	  s = Sb = Ss = lens = 0;
 	  for(i = 0; i < overlap; i++)
 	    {
 	      if(new[lastscan + lenf - overlap + i] == old[lastpos + lenf - overlap + i])
 		s++;
 	      if(new[scan - lenb + i] == old[pos - lenb + i])
-		s--;
-	      if (s > Ss)
+		Sb++;
+	      if (s > Sb && s - Sb > Ss)
 		{
-		  Ss = s;
+		  Ss = s - Sb;
 		  lens = i + 1;
 		}
 	    }
@@ -887,7 +955,7 @@ void mkdiff(int mode, unsigned char *old, unsigned int oldlen, unsigned char *ne
 	  instr[instrlen].copyout = lenf;
 	  instr[instrlen].copyin = (scan - lenb) - (lastscan + lenf);
 	  instr[instrlen].copyinoff = lastscan + lenf;
-	  instr[instrlen].seekout = (pos - lenb) - (lastpos + lenf);
+	  instr[instrlen].copyoutoff = lastpos;
 	  instrlen++;
 	}
       if (bzi)
@@ -897,29 +965,45 @@ void mkdiff(int mode, unsigned char *old, unsigned int oldlen, unsigned char *ne
 	  addoff(bzi, (pos - lenb) - (lastpos + lenf));
 	}
       if (bze)
-	blockwrite(bze, new + lastscan + lenf, (scan - lenb) - (lastscan + lenf));
-      while (lenf > 0)
 	{
-	  unsigned char addblk[4096];
-	  int len;
-	  len = lenf > 4096 ? 4096 : lenf;
-	  for (i = 0; i < len; i++)
-	    addblk[i] = new[lastscan + i] - old[lastpos + i];
-	  if (blockwrite(bza, addblk, len) != len)
+	  i = lastscan + lenf;
+	  s = (scan - lenb) - i;
+	  while (s > 0)
 	    {
-	      fprintf(stderr, "could not append to data block\n");
-	      exit(1);
+	      int len2 = s > 0x40000000 ? 0x40000000 : s;
+	      blockwrite(bze, new + i, len2);
+	      i += len2;
+	      s -= len2;
 	    }
-	  lastscan += len;
-	  lastpos += len;
-	  lenf -= len;
 	}
+      if (!bza)
+	{
+	  lastscan += lenf;
+	  lastpos += lenf;
+	}
+      else
+	while (lenf > 0)
+	  {
+	    unsigned char addblk[4096];
+	    int len2;
+	    len2 = lenf > 4096 ? 4096 : lenf;
+	    for (i = 0; i < len2; i++)
+	      addblk[i] = new[lastscan + i] - old[lastpos + i];
+	    if (blockwrite(bza, addblk, len2) != len2)
+	      {
+		fprintf(stderr, "could not append to data block\n");
+		exit(1);
+	      }
+	    lastscan += len2;
+	    lastpos += len2;
+	    lenf -= len2;
+	  }
 
       lastscan = scan - lenb;
       lastpos = pos - lenb;
       lastoffset = pos - scan;
     }
-  if (blockclose(bza, addblkp, addblklenp))
+  if (bza && blockclose(bza, addblkp, addblklenp))
     {
       fprintf(stderr, "could not close data block\n");
       exit(1);
