@@ -21,25 +21,6 @@
  *
  */
 
-static ssize_t
-xread(int fd, void *buf, size_t count)
-{
-  size_t c = 0;
-  ssize_t r;
-  while (count > 0)
-    {
-      r = read(fd, buf, count);
-      if (r < 0)
-	return r;
-      if (r == 0)
-        return c;
-      buf += r;
-      c += r;
-      count -= r;
-    }
-  return c;
-}
-
 struct rpmhead *
 readhead(int fd, int pad)
 {
@@ -78,6 +59,40 @@ readhead(int fd, int pad)
   return h;
 }
 
+struct rpmhead *
+readhead_buf(unsigned char *buf, int len, int pad)
+{
+  int cnt, dcnt;
+  struct rpmhead *h;
+
+  if (len < 16)
+    {
+      fprintf(stderr, "bad header\n");
+      return 0;
+    }
+  if (buf[0] != 0x8e || buf[1] != 0xad || buf[2] != 0xe8 || buf[3] != 0x01)
+    {
+      fprintf(stderr, "bad header\n");
+      return 0;
+    }
+  cnt  = buf[8] << 24  | buf[9] << 16  | buf[10] << 8 | buf[11];
+  dcnt = buf[12] << 24 | buf[13] << 16 | buf[14] << 8 | buf[15];
+  if ((dcnt & 7) && pad)
+    dcnt += 8 - (dcnt & 7);
+  if (len < 16 + cnt * 16 + dcnt)
+    {
+      fprintf(stderr, "bad header\n");
+      return 0;
+    }
+  h = xmalloc(sizeof(*h) + cnt * 16 + dcnt);
+  memcpy(h->intro, buf, 16);
+  memcpy(h->data, buf + 16, cnt * 16 + dcnt);
+  h->cnt = cnt;
+  h->dcnt = dcnt;
+  h->dp = h->data + cnt * 16;
+  return h;
+}
+
 unsigned int *
 headint32(struct rpmhead *h, int tag, int *cnt)
 {
@@ -101,7 +116,7 @@ headint32(struct rpmhead *h, int tag, int *cnt)
   if (o + 4 * i > h->dcnt)
     return 0;
   d = h->dp + o;
-  r = xmalloc(sizeof(unsigned int) * (i ? i : 1));
+  r = xmalloc2(i ? i : 1, sizeof(unsigned int));
   if (cnt)
     *cnt = i;
   for (o = 0; o < i; o++, d += 4)
@@ -132,7 +147,7 @@ headint16(struct rpmhead *h, int tag, int *cnt)
   if (o + 2 * i > h->dcnt)
     return 0;
   d = h->dp + o;
-  r = xmalloc(sizeof(unsigned int) * (i ? i : 1));
+  r = xmalloc2(i ? i : 1, sizeof(unsigned int));
   if (cnt)
     *cnt = i;
   for (o = 0; o < i; o++, d += 2)
@@ -182,7 +197,7 @@ headstringarray(struct rpmhead *h, int tag, int *cnt)
     return 0;
   o = d[8] << 24 | d[9] << 16 | d[10] << 8 | d[11];
   i = d[12] << 24 | d[13] << 16 | d[14] << 8 | d[15];
-  r = xmalloc(sizeof(char *) * (i ? i : 1));
+  r = xmalloc2(i ? i : 1, sizeof(char *));
   if (cnt)
     *cnt = i;
   d = h->dp + o;
@@ -200,6 +215,30 @@ headstringarray(struct rpmhead *h, int tag, int *cnt)
   return r;
 }
 
+unsigned char *
+headbin(struct rpmhead *h, int tag, int len)
+{
+  unsigned int i, o;
+  unsigned char *d, taga[4];
+  d = h->data;
+  taga[0] = tag >> 24;
+  taga[1] = tag >> 16;
+  taga[2] = tag >> 8;
+  taga[3] = tag;
+  for (i = 0; i < h->cnt; i++, d += 16)
+    if (d[3] == taga[3] && d[2] == taga[2] && d[1] == taga[1] && d[0] == taga[0])
+      break;
+  if (i >= h->cnt)
+    return 0;
+  if (d[4] != 0 || d[5] != 0 || d[6] != 0 || d[7] != 7)
+    return 0;
+  i = d[12] << 24 | d[13] << 16 | d[14] << 8 | d[15];
+  if (len != i)
+    return 0;
+  o = d[8] << 24 | d[9] << 16 | d[10] << 8 | d[11];
+  return (unsigned char *)h->dp + o;
+}
+
 char **
 headexpandfilelist(struct rpmhead *h, int *cnt)
 {
@@ -240,26 +279,34 @@ headexpandfilelist(struct rpmhead *h, int *cnt)
 char *headtonevr(struct rpmhead *h)
 {
   char *name;
-  char *epoch; 
+  unsigned int *epoch; 
   char *version;
   char *release;
   char *nevr;
+  int epochcnt = 0;
 
   name = headstring(h, TAG_NAME);
   version  = headstring(h, TAG_VERSION);
   release  = headstring(h, TAG_RELEASE);
-  epoch = headstring(h, TAG_EPOCH);
+  epoch = headint32(h, TAG_EPOCH, &epochcnt);
   if (!name || !version || !release)
     {
       fprintf(stderr, "headtonevr: bad rpm header\n");
       exit(1);
     }
-  if (epoch && *epoch == 0)
-    epoch = 0;
-  nevr = xmalloc(strlen(name) + 1 + strlen(version) + 1 + strlen(release) + 1 + (epoch ? strlen(epoch + 1) : 0));
-  if (epoch) 
-    sprintf(nevr, "%s-%s:%s-%s", name, epoch, version, release);
+  if (epoch && epochcnt)
+    {
+      char epochbuf[11];	/* 32bit decimal will fit in */
+      sprintf(epochbuf, "%u", *epoch);
+      nevr = xmalloc(strlen(name) + 1 + strlen(epochbuf) + 1 + strlen(version) + 1 + strlen(release) + 1);
+      sprintf(nevr, "%s-%s:%s-%s", name, epochbuf, version, release);
+    }
   else
-    sprintf(nevr, "%s-%s-%s", name, version, release);
+    {
+      nevr = xmalloc(strlen(name) + 1 + strlen(version) + 1 + strlen(release) + 1);
+      sprintf(nevr, "%s-%s-%s", name, version, release);
+    }
+  if (epoch)
+    free(epoch);
   return nevr;
 }
